import os

import boto3
import numpy as np
from openai import OpenAI

from contextlib import closing

rng = np.random.default_rng()

# picking neural voices
VOICE_MAP = {
    ("english", "female"): "Vitoria",
    ("english", "male"): "Gregory",
    ("portuguese", "female"): "Camila",
    ("portuguese", "male"): "Thiago",
    ("spanish", "female"): "Lucia",
    ("spanish", "male"): "Sergio",
    ("italian", "female"): "Bianca",
    ("italian", "male"): "Adriano",
}


class ToolTextToSpeech:
    def __init__(self):
        self.name = "text_to_speech"

        self.tool_description = {
            "name": self.name,
            "description": """Uses a text-to-speech tool to convert the provided text into spoken audio in the given language.
You need to specify the text that will be spoken, the language and the speaker gender. The language is used to select the appropriate text-to-speech model.

The user can listen to the audio without exposing the auto-generated file names. Do NOT include actual file names in the answer. Do NOT include the <path_to_audio></path_to_audio> tag in the answer. Only mention that the image has been generated successfully.

Raises ValueError: if not able to generate the audio.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "input_text": {
                        "type": "string",
                        "description": "Text that will be converted to speech.",
                    },
                    "language": {
                        "type": "string",
                        "enum": ["english", "portuguese", "spanish", "italian"],
                        "description": "Language of the input_text. Must be one of english, portuguese, spanish or italian",
                    },
                    "speaker_gender": {
                        "type": "string",
                        "enum": ["male", "female"],
                        "description": "Desired gender of the speaker. Default is female. Can be male or female.",
                    },
                    "tts_engine": {
                        "type": "string",
                        "enum": ["aws_polly", "openai"],
                        "description": """Must be aws_polly or openai.
Select the neural engine that will be used to generate the audio. Defaults to OpenAI text to speech. If not available, you must manually select the other.""",
                    },
                    "instructions": {
                        "type": "string",
                        "description": """Use ONLY if tts_engine = openai.
Optional parameter to control the voice of the generated audio with additional <aspects_of_speech></aspects_of_speech>.
<aspects_of_speech>
Accent
Emotional range
Intonation
Impressions
Speed of speech
Tone
Whispering
</aspects_of_speech>
You can prompt the model to control any combination of the <aspects_of_speech></aspects_of_speech> by giving a proper description for each item of interest.
""",
                    },
                },
                "required": ["input_text", "language"],
            },
        }
        self.polly_client = None
        self.openai_client = None

    def __call__(
        self,
        input_text,
        language,
        speaker_gender="female",
        tts_engine="openai",
        instructions="",
        **kwargs,
    ):
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        # pick TTS engine
        tts_engine = tts_engine.lower().strip()
        valid_tts_engines = ["aws_polly", "openai"]
        assert (
            tts_engine in valid_tts_engines
        ), f"Invalid text to speech engine: {tts_engine}. Must be one of {valid_tts_engines}."
        if tts_engine == "openai" and self.openai_client is None:
            self.openai_client = OpenAI()
        elif tts_engine == "aws_polly" and self.polly_client is None:
            self.polly_client = boto3.client(
                service_name="polly", region_name="us-west-2"
            )

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/audio_{rng_num}.mp3"

        try:
            if tts_engine == "aws_polly":
                # Request speech synthesis
                response = self.polly_client.synthesize_speech(
                    Text=input_text,
                    OutputFormat="mp3",
                    VoiceId=VOICE_MAP[
                        (language.lower(), speaker_gender.lower())
                    ],  # "Danielle",
                    Engine="neural",
                )
                if "AudioStream" in response:
                    with closing(response["AudioStream"]) as stream:
                        with open(target_file, "wb") as file:
                            file.write(stream.read())
            elif tts_engine == "openai":
                response = self.openai_client.audio.speech.create(
                    model="gpt-4o-mini-tts",
                    voice="onyx" if speaker_gender == "male" else "nova",
                    input=input_text,
                    instructions=instructions,
                )

                response.stream_to_file(target_file)
        except Exception as e:
            return f"Audio was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Audio was not saved correctly."

        ans = ["<audio>"]
        ans.append(f"<path_to_audio>{target_file}</path_to_audio>")
        ans.append("</audio>")
        return "\n".join(ans)



import pytest
from unittest.mock import patch, Mock
from gat_llm.tools.make_qr_code import ToolMakeQRCode


@pytest.fixture
def mock_qrcode():
    with patch("qrcode.QRCode") as mock:
        mock_qr = Mock()
        mock_qr.make_image.return_value = Mock()
        mock.return_value = mock_qr
        yield mock


@pytest.fixture
def mock_os_path_isfile():
    with patch("os.path.isfile") as mock:
        mock.return_value = True
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    tmqc = ToolMakeQRCode()
    ans = tmqc("Test QR Code", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_make_qr_code_success(mock_qrcode, mock_os_path_isfile):
    tmqc = ToolMakeQRCode()
    result = tmqc("Test QR Code")
    assert "<image>" in result
    assert "<path_to_image>" in result
    assert "</image>" in result


def test_make_qr_code_with_error_correction(mock_qrcode, mock_os_path_isfile):
    tmqc = ToolMakeQRCode()
    result = tmqc("Test QR Code", error_correction="high")
    assert "<image>" in result
    assert "<path_to_image>" in result
    assert "</image>" in result


def test_make_qr_code_invalid_error_correction():
    tmqc = ToolMakeQRCode()
    result = tmqc("Test QR Code", error_correction="invalid")
    assert "error_correction must be one of" in result


@patch("os.path.isfile", return_value=False)
def test_make_qr_code_file_not_saved(mock_isfile, mock_qrcode):
    tmqc = ToolMakeQRCode()
    result = tmqc("Test QR Code")
    assert "Error: Image was not saved correctly" in result


# Add more tests for different scenarios and edge cases



import pytest

from gat_llm.tools.summarize_past import ToolSummarizePast


def test_unexpected_arg(unexpected_param_msg):
    tsp = ToolSummarizePast()
    ans = tsp("past", "info", "next", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


@pytest.mark.parametrize("past_conversation_summary", ["past1", "past2", "past3"])
@pytest.mark.parametrize("information_still_relevant", ["info1", "info2", "info3"])
@pytest.mark.parametrize("next_question_to_answer", ["next1", "next2"])
def test_complete_ans(
    past_conversation_summary,
    information_still_relevant,
    next_question_to_answer,
):
    """Checks if the answer actually contains the arguments
    (expected for this function)
    """
    tsp = ToolSummarizePast()
    ans = tsp(
        past_conversation_summary,
        information_still_relevant,
        next_question_to_answer,
    )
    assert past_conversation_summary in ans
    assert information_still_relevant in ans
    assert next_question_to_answer in ans



import os


class ToolReadLocalFolder:
    def __init__(self):
        self.name = "read_file_names_in_local_folder"

        self.tool_description = {
            "name": self.name,
            "description": """Reads the file names contained in a local folder.

Raises ValueError: if the folder does not exist.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "path_to_folder": {
                        "type": "string",
                        "description": """Local path to the files whose contents should be retrieved. Provide one file per line as in the <example></example>:
<example>
file1.txt
file2.py
subfolder/file3.docx
</example>""",
                    },
                },
                "required": ["path_to_folder"],
            },
        }

    def __call__(self, path_to_folder, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        if not os.path.isdir(path_to_folder):
            return f"Error: Did not find folder `{path_to_folder}`"

        ans = [os.path.join(path_to_folder, x) for x in os.listdir(path_to_folder)]
        ans = [x for x in ans if os.path.isfile(x)]

        return f"<files><file>{'</file><file>'.join(ans)}</file></files>"



import subprocess


class ToolRunWithPython:
    def __init__(self):
        self.name = "run_with_python"

        self.tool_description = {
            "name": self.name,
            "description": """Changes to target folder and runs a python file with the command line `python <file_name.py>`.
Note that the script will be run in the `test_env` environment. The issued command will be `cd && conda activate test_env && python file_name`.
If the environment does not exist or a package is missing, instruct the user about how to fix it.

This tool returns python_stdout containing the stdout of python with errors returned.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "file_name": {
                        "type": "string",
                        "description": "Name of the Python file to run.",
                    },
                    "exec_folder": {
                        "type": "string",
                        "description": "Folder to change to before executing the Python script.",
                    },
                },
                "required": ["file_name", "exec_folder"],
            },
        }

    def __call__(self, file_name, exec_folder, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        all_args = f"cd {exec_folder} && conda activate test_env && python {file_name}"
        p = subprocess.run(all_args, capture_output=True, text=True, shell=True)

        final_ans = ["<python_stdout>"]
        final_ans.append(p.stdout + p.stderr)
        final_ans.append("</python_stdout>")
        return "\n".join(final_ans)



import pytest
from unittest.mock import patch
from gat_llm.tools.solve_numeric import ToolSolveNumeric


def test_unexpected_arg(unexpected_param_msg):
    tsn = ToolSolveNumeric()
    ans = tsn("import numpy as np", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_solve_numeric_success():
    tsn = ToolSolveNumeric()
    numpy_code = """
import numpy as np
arr = np.array([1, 2, 3, 4, 5])
ans = np.mean(arr)
"""
    result = tsn(numpy_code)
    assert result == "3.0"


def test_solve_numeric_execution_error():
    tsn = ToolSolveNumeric()
    numpy_code = "invalid_code"
    result = tsn(numpy_code)
    assert "Code did NOT execute correctly" in result
    assert "Error description" in result


@pytest.mark.parametrize(
    "code,expected",
    [
        ("import numpy as np\nans = np.sum([1, 2, 3, 4, 5])", "15"),
        ("import numpy as np\nans = np.prod([1, 2, 3, 4, 5])", "120"),
        ("import numpy as np\nans = np.max([1, 2, 3, 4, 5])", "5"),
        ("import numpy as np\nans = np.min([1, 2, 3, 4, 5])", "1"),
    ],
)
def test_solve_numeric_various_operations(code, expected):
    tsn = ToolSolveNumeric()
    result = tsn(code)
    assert result == expected


# Add more tests for different scenarios and edge cases



import os
import numpy as np


ans = None


class ToolSolveNumeric:
    def __init__(self):
        self.name = "solve_numeric"

        self.tool_description = {
            "name": self.name,
            "description": f"""Evaluates numerical expressions the Python library numpy. Always use {self.name} to evaluate the numerical expressions instead of doing it manually. Make sure to import numpy before using it in the numpy_code.
Ensure that the code generates the final answer to the problem, without requiring any further analysis. Assign the final answer to the variable "ans".

Use {self.name} in these <use_cases></use_cases>:
<use_cases>
<use_case>Doing direct numeric evaluation of expressions</use_case>
<use_case>There are no symbolic equations to solve</use_case>
</use_cases>

Raises ValueError: if the code to be executed was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "numpy_code": {
                        "type": "string",
                        "description": "Python code that, when executed, will provide the final answer to the problem",
                    },
                },
                "required": ["numpy_code"],
            },
        }

    def __call__(self, numpy_code, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        try:
            global ans
            if ans is not None:
                ans = None
            exec(numpy_code, globals())
        except Exception as e:
            return f"Code did NOT execute correctly.\nError description: {str(e)}"

        return str(ans)



import pytest
from unittest.mock import patch, Mock
import pandas as pd
from gat_llm.tools.query_database import ToolQueryLLMDB, SampleOrder_LLM_DB


@pytest.fixture
def mock_duckdb_sql():
    with patch("duckdb.sql") as mock:
        mock.return_value.df = Mock(return_value=pd.DataFrame({"col1": range(10)}))
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    db = SampleOrder_LLM_DB()
    tqd = ToolQueryLLMDB(db)
    ans = tqd("SELECT * FROM tblSales", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_query_database_success(mock_duckdb_sql):
    db = SampleOrder_LLM_DB()
    tqd = ToolQueryLLMDB(db)
    result = tqd("SELECT * FROM tblSales LIMIT 5")
    assert "SQL code executed correctly" in result
    assert "<query_results>" in result
    assert "</query_results>" in result


def test_query_database_too_many_records(mock_duckdb_sql):
    db = SampleOrder_LLM_DB()
    tqd = ToolQueryLLMDB(db, max_records=5)
    mock_duckdb_sql.df = pd.DataFrame({"col1": range(10)})
    result = tqd("SELECT * FROM tblSales")
    assert "SQL code NOT executed. Too many records" in result
    assert "Number of records found: 10" in result


def test_query_database_error(mock_duckdb_sql):
    db = SampleOrder_LLM_DB()
    tqd = ToolQueryLLMDB(db)
    with patch("duckdb.sql", side_effect=Exception("SQL error")):
        result = tqd("INVALID SQL")
    assert "SQL code NOT executed. Error description" in result
    assert "SQL error" in result


# Add more tests for different scenarios and edge cases



ans = None


class ToolSolveSymbolic:
    def __init__(self):
        self.name = "solve_symbolic"

        self.tool_description = {
            "name": self.name,
            "description": f"""Gets the solution to a symbolic mathematics problem using the Python library sympy. Make sure to import sympy before using it in the <sympy_code></sympy_code>.
Ensure that the code generates the final answer to the problem, without requiring any further analysis. Assign the final answer to the variable "ans".

Use {self.name} only in these <use_cases>:
<use_cases>
<use_case>There are symbolic equations or expressions to simplify or solve</use_case>
<use_case>The problem cannot be solved using purely numeric evaluations</use_case>
</use_cases>

Before calling this function, explicitly list out all problem assumptions in the <scratchpad></scratchpad>.

Raises ValueError: if the code to be executed was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "sympy_code": {
                        "type": "string",
                        "description": "Python code that, when executed, will provide the final answer to the problem",
                    },
                },
                "required": ["sympy_code"],
            },
        }

    def __call__(self, sympy_code, **kwargs):
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        try:
            global ans
            if ans is not None:
                ans = None
            exec(sympy_code, globals())
        except Exception as e:
            return f"Code did NOT execute correctly.\nError description: {str(e)}"

        return str(ans)



import pytest
from unittest.mock import patch, Mock
from gat_llm.tools.use_ffmpeg import ToolUseFFMPEG


@pytest.fixture
def mock_subprocess_run():
    with patch("subprocess.run") as mock:
        mock_process = Mock()
        mock_process.stdout = "ffmpeg output"
        mock_process.stderr = ""
        mock.return_value = mock_process
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    tuf = ToolUseFFMPEG()
    ans = tuf("-i input.mp4 output.mp4", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_use_ffmpeg_success(mock_subprocess_run):
    tuf = ToolUseFFMPEG()
    result = tuf("-i input.mp4 output.mp4")
    assert "<ffmpeg_stdout>" in result
    assert "ffmpeg output" in result
    assert "</ffmpeg_stdout>" in result
    mock_subprocess_run.assert_called_once_with(
        "ffmpeg -i input.mp4 output.mp4", capture_output=True, text=True, shell=True
    )


def test_use_ffmpeg_with_error(mock_subprocess_run):
    mock_subprocess_run.return_value.stderr = "ffmpeg error"
    tuf = ToolUseFFMPEG()
    result = tuf("-i nonexistent.mp4 output.mp4")
    assert "<ffmpeg_stdout>" in result
    assert "ffmpeg error" in result
    assert "</ffmpeg_stdout>" in result


@pytest.mark.parametrize(
    "ffmpeg_args",
    [
        "-i input.mp4 -ss 10 -t 20 output.mp4",
        '-i input.mp4 -filter_complex "[0:v]setpts=0.5*PTS[v];[0:a]atempo=2.0[a]" -map "[v]" -map "[a]" output.mp4',
        "-i input.mp4 -vf scale=1280:720 output.mp4",
    ],
)
def test_use_ffmpeg_various_commands(ffmpeg_args, mock_subprocess_run):
    tuf = ToolUseFFMPEG()
    result = tuf(ffmpeg_args)
    assert "<ffmpeg_stdout>" in result
    assert "ffmpeg output" in result
    assert "</ffmpeg_stdout>" in result
    mock_subprocess_run.assert_called_once_with(
        f"ffmpeg {ffmpeg_args}", capture_output=True, text=True, shell=True
    )


# Add more tests for different scenarios and edge cases



import pytest
from unittest.mock import patch, mock_open
from gat_llm.tools.update_user_details import ToolUpdateUserDetails


@pytest.fixture
def mock_os_path_isfile():
    with patch("os.path.isfile") as mock:
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    tuud = ToolUpdateUserDetails()
    ans = tuud("READ", username="testuser", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_read_user_details_success(mock_os_path_isfile):
    mock_os_path_isfile.return_value = True
    tuud = ToolUpdateUserDetails()
    with patch(
        "builtins.open",
        mock_open(read_data="[user_info]\n[name]John Doe[/name]\n[/user_info]"),
    ):
        result = tuud("READ", username="testuser")
    assert result == "[user_info]\n[name]John Doe[/name]\n[/user_info]"


def test_read_user_details_not_found(mock_os_path_isfile):
    mock_os_path_isfile.return_value = False
    tuud = ToolUpdateUserDetails()
    result = tuud("READ", username="testuser")
    assert result == "No information available."


def test_write_user_details_success():
    tuud = ToolUpdateUserDetails()
    with patch("builtins.open", mock_open()) as mock_file:
        result = tuud(
            "WRITE",
            username="testuser",
            contents="[user_info]\n[name]Jane Doe[/name]\n[/user_info]",
        )
    assert result == "User information saved successfully."
    mock_file.assert_called_once_with("user_info/testuser.txt", "w", encoding="utf-8")
    mock_file().write.assert_called_once_with(
        "[user_info]\n[name]Jane Doe[/name]\n[/user_info]"
    )


def test_invalid_action():
    tuud = ToolUpdateUserDetails()
    result = tuud("INVALID", username="testuser")
    assert "Action must be one of" in result


# Add more tests for different scenarios and edge cases



import pytest

from gat_llm.tools.do_date_math import ToolDoDateMath


@pytest.mark.parametrize(
    "initial_date,delta,delta_type,expected",
    [
        ("2024-10-01", "1", "day", "2024-10-02 Wednesday"),
        ("2024-10-01", "2, 4", "day", "2024-10-03 Thursday,2024-10-05 Saturday"),
        ("2024-10-03", "-1", "day", "2024-10-02 Wednesday"),
        ("2024-10-09", "0", "day", "2024-10-09 Wednesday"),
        ("2024-10-09", "0", "month", "2024-10-09 Wednesday"),
        ("2024-10-09", "0", "year", "2024-10-09 Wednesday"),
        ("2023-10-01", "2", "month", "2023-12-01 Friday"),
        ("2023-10-01", "3,-3", "month", "2024-01-01 Monday,2023-07-01 Saturday"),
        (
            "2023-10-01",
            "1,-1,0, -3,3, 4",
            "year",
            "2024-10-01 Tuesday,2022-10-01 Saturday,2023-10-01 Sunday,2020-10-01 Thursday,2026-10-01 Thursday,2027-10-01 Friday",
        ),
    ],
)
def test_sum_periods(initial_date, delta, delta_type, expected):
    dm = ToolDoDateMath()
    computed_date = dm(initial_date, delta, delta_type)
    assert computed_date == expected


def test_unexpected_arg(unexpected_param_msg):
    dm = ToolDoDateMath()
    computed_date = dm("initial_date", "delta", "delta_type", unexpected_argument=None)
    assert computed_date == f"{unexpected_param_msg}unexpected_argument"



import os
import numpy as np
from matplotlib import pyplot as plt

rng = np.random.default_rng()


class ToolMakeCustomPlot:
    def __init__(self):
        self.name = "make_custom_plot"

        self.save_code = "plt.savefig('media/plot.jpg')"

        self.tool_description = {
            "name": self.name,
            "description": f"""Generates an image or plot using the provided custom Python code plot_code, whose only dependencies should be numpy and matplotlib. This tool is useful when the user requests plots to be generated from data that was previously retrieved.

If the user didn't specifically ask for a plot, confirm is the user wants to generate a plot or visualization before calling this tool.

The last line of the plot code must save the figure to 'media/plot.jpg'. The last line of the code should be: {self.save_code}

The user can view the images without exposing the auto-generated file names. Do NOT include actual file names in the answer. Do NOT include the <path_to_image></path_to_image> tag in the answer. Only mention that the image has been generated successfully.

Raises ValueError: if the code to be executed was invalid.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "plot_code": {
                        "type": "string",
                        "description": "Python code that, when executed, will generate an image or plot requested by the user",
                    },
                },
                "required": ["plot_code"],
            },
        }

    def __call__(self, plot_code, **kwargs):
        os.makedirs("media", exist_ok=True)
        # fix weird save attempts
        plot_code = plot_code.splitlines()
        plot_code = [x for x in plot_code if not x.startswith("plt.savefig")]
        plot_code.append(self.save_code)
        plot_code = "\n".join(plot_code)

        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"'media/plot_{rng_num}.jpg'"
        plot_code = plot_code.replace("'media/plot.jpg'", target_file)

        try:
            exec(plot_code)
        except Exception as e:
            return f"Plot was NOT generated.\nError description: {str(e)}"

        target_file = target_file.replace("'", "")
        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        # with open(f'plot_code_{rng_num}.txt', 'w', encoding='utf-8') as debug_file:
        #    p_str = plot_code + f'\n***{target_file} {os.path.isfile(target_file)}'
        #    debug_file.write(p_str)

        ans = ["<image>"]
        ans.append(f"<path_to_image>{target_file}</path_to_image>")
        ans.append("</image>")
        return "\n".join(ans)



import os
import json
import boto3
import base64

import numpy as np
from openai import OpenAI

rng = np.random.default_rng()


class ToolTextToImage:
    def _gen_img_bedrock(self, input_text, target_file, default_region="us-west-2"):
        if self.bedrock_client is None:
            self.bedrock_client = boto3.client(
                "bedrock-runtime", region_name=default_region
            )
        model_id = "stability.stable-image-ultra-v1:0"

        # seed = int(rng.integers(low=0, high=4294967295))
        native_request = {
            "prompt": input_text,
        }
        request = json.dumps(native_request)
        response = self.bedrock_client.invoke_model(modelId=model_id, body=request)
        output_body = json.loads(response["body"].read().decode("utf-8"))
        base64_output_image = output_body["images"][0]

        # Extract the image data.
        image_data = base64.b64decode(base64_output_image)
        with open(target_file, "wb") as file:
            file.write(image_data)
        return input_text

    def _gen_img_openai(self, input_text, target_file):
        if self.openai_client is None:
            self.openai_client = OpenAI()
        response = self.openai_client.images.generate(
            model="gpt-image-1",
            prompt=input_text,
            n=1,
            moderation="low",
            output_format="png",
        )
        parsed_ans = json.loads(response.json())
        revised_prompt = parsed_ans["data"][0]["revised_prompt"]
        img_content = base64.b64decode(parsed_ans["data"][0]["b64_json"])
        with open(target_file, "wb") as f:
            f.write(img_content)
        return revised_prompt

    def __init__(self):
        self.name = "text_to_image"

        self.valid_engines = ["openai-img", "bedrock-stablediffusion"]
        self.tool_description = {
            "name": self.name,
            "description": """Uses a text-to-image tool to convert the provided text into an image. Uses neural methods such as diffusion.
You need to specify the text that will be converted to the image.

Raises ValueError: if not able to generate the image.""",
            "input_schema": {
                "type": "object",
                "properties": {
                    "input_text": {
                        "type": "string",
                        "description": "Text that will be converted to image.",
                    },
                    "engine": {
                        "type": "string",
                        "description": f"Engine that will be used to generate the image. Valid values are: {self.valid_engines}. Default value is {self.valid_engines[0]}",
                    },
                },
                "required": ["input_text"],
            },
        }
        self.openai_client = None
        self.bedrock_client = None

    def __call__(
        self,
        input_text,
        engine=None,
        **kwargs,
    ):
        os.makedirs("media", exist_ok=True)
        if len(kwargs) > 0:
            return f"Error: Unexpected parameter(s): {','.join([x for x in kwargs])}"

        rng_num = rng.integers(low=0, high=900000)
        target_file = f"media/gen_image_{rng_num}.png"

        if engine is None:
            if os.getenv("OPENAI_API_KEY"):
                engine = "openai-img"
            else:
                engine = "bedrock-stablediffusion"
        engine = engine.lower()

        if engine not in self.valid_engines:
            return f"Invalid text-to-image engine: {engine}. Valid engines are: {self.valid_engines}"

        try:
            if engine == "openai-img":
                revised_prompt = self._gen_img_openai(input_text, target_file)
            elif engine == "bedrock-stablediffusion":
                revised_prompt = self._gen_img_bedrock(input_text, target_file)
        except Exception as e:
            return f"Image was NOT generated.\nError description: {str(e)}"

        if not os.path.isfile(target_file):
            return "Error: Image was not saved correctly."

        ans = ["<image>"]
        ans.append(
            f"<revised_prompt>{revised_prompt}</revised_prompt><used_engine>{engine}</used_engine><path_to_image>{target_file}</path_to_image>"
        )
        ans.append("</image>")
        return "\n".join(ans)



import time
import xml.etree.ElementTree as ET

import numpy as np

from .make_custom_plot import ToolMakeCustomPlot
from .solve_symbolic import ToolSolveSymbolic
from .solve_numeric import ToolSolveNumeric
from .solve_python_code import ToolSolvePythonCode
from .get_webpage_contents import ToolGetUrlContent
from .make_qr_code import ToolMakeQRCode
from .read_local_file import ToolReadLocalFile
from .write_local_file import ToolWriteLocalFile
from .read_file_names_in_local_folder import ToolReadLocalFolder
from .do_date_math import ToolDoDateMath
from .update_user_details import ToolUpdateUserDetails
from .use_ffmpeg import ToolUseFFMPEG
from .plot_with_graphviz import ToolPlotWithGraphviz
from .text_to_speech import ToolTextToSpeech
from .speech_to_text import ToolSpeechToText
from .summarize_past import ToolSummarizePast
from .text_to_image import ToolTextToImage
from .image_edit import ToolImageEdit
from .query_database import ToolQueryLLMDB
from .query_database import SampleOrder_LLM_DB
from .run_with_python import ToolRunWithPython
from .select_video_frames import ToolSelectVideoFrames

rng = np.random.default_rng()


class LLMTools:
    def get_all_tools():
        """Returns a list of all tools available"""
        return [
            ToolDoDateMath(),
            ToolUpdateUserDetails(),
            ToolMakeCustomPlot(),
            ToolSolveSymbolic(),
            ToolSolveNumeric(),
            ToolGetUrlContent(None),
            ToolMakeQRCode(),
            ToolReadLocalFile(),
            ToolWriteLocalFile(),
            ToolReadLocalFolder(),
            ToolUseFFMPEG(),
            ToolSelectVideoFrames(),
            ToolSolvePythonCode(),
            ToolPlotWithGraphviz(),
            ToolTextToSpeech(),
            ToolSpeechToText(),
            ToolTextToImage(),
            ToolImageEdit(),
            ToolSummarizePast(),
            ToolQueryLLMDB(SampleOrder_LLM_DB()),
            ToolRunWithPython(),
        ]

    def __init__(self, query_llm=None, desired_tools=None):
        """Constructor."""
        self.query_llm = query_llm
        if desired_tools is None:
            self.tools = [
                ToolDoDateMath(),
                ToolMakeCustomPlot(),
                ToolSolveSymbolic(),
                ToolSolveNumeric(),
                ToolGetUrlContent(self.query_llm),
                ToolMakeQRCode(),
                ToolReadLocalFile(),
                ToolWriteLocalFile(),
                ToolReadLocalFolder(),
                ToolUseFFMPEG(),
                ToolSelectVideoFrames(),
                ToolPlotWithGraphviz(),
                ToolRunWithPython(),
                # Being left out for now. Just uncomment to enable
                # ToolSummarizePast(),
                # ToolUpdateUserDetails(),
                # ToolQueryLLMDB(SampleOrder_LLM_DB()),
                # ToolSolvePythonCode(),
                ToolTextToSpeech(),
                ToolSpeechToText(),
                ToolTextToImage(),
                ToolImageEdit(),
            ]
        else:
            self.tools = desired_tools

        self.tool_mapping = dict(zip([x.name for x in self.tools], self.tools))
        self.call_return_string = """
<function_results>
<result>
<tool_name>{{TOOLNAME}}</tool_name>
<stdout>
{{TOOLRESULTS}}
</stdout>
</result>
</function_results>
<scratchpad> To answer the question, I still need to:"""
        self.invoke_log = []

    def invoke_from_cmd(self, xml_cmd, username=None):
        """Invokes a tool given the xml command sent by the LLM"""
        cmd = self.parse_command(xml_cmd)
        cmd["parameters"]["username"] = str(username)
        return self.invoke_tool(cmd["tool_name"], **cmd["parameters"])

    def parse_command(self, xml_cmd):
        """Parses a XML command to retrieve arguments and tool name"""
        try:
            assert (
                "&" not in xml_cmd
            ), "Do not include the character & in any of the arguments."

            root = ET.fromstring(xml_cmd)
            ans = {}
            func_params = {}
            for child in root:
                if child.tag == "parameters":
                    for x in child:
                        func_params[x.tag] = x.text
                if child.tag == "tool_name":
                    ans["tool_name"] = child.text
            ans["parameters"] = func_params
            return ans
        except Exception as e:
            print(str(e))
            return {"tool_name": "unknown", "parameters": {}, "error": str(e)}

    def get_tool_descriptions(self):
        """Retrieves the description of all tools available"""
        desc_list = ["<tools>"]
        for x in self.tools:
            desc_list.append(self._parse_tool_description(x.tool_description))
        desc_list.append("</tools>")
        return "\n".join(desc_list)

    def _parse_tool_description(self, description_dict):
        """Given a tool description in JSON, parse its contents"""
        textual_desc = {}
        textual_desc["tool_name"] = description_dict["name"]
        textual_desc["description"] = description_dict["description"]

        # parameters
        textual_desc["parameters"] = []
        cur_desc_dict = description_dict["input_schema"]["properties"]
        for k in cur_desc_dict:
            cur_param = {"parameter": {}}
            cur_param["parameter"] = {"name": k}
            for k2 in cur_desc_dict[k]:
                cur_param["parameter"][k2] = str(cur_desc_dict[k][k2])
            textual_desc["parameters"].append(self._json2xml(cur_param))
        textual_desc["required_parameters"] = str(
            description_dict["input_schema"]["required"]
        )

        return self._json2xml({"tool_description": textual_desc})

    def _json2xml(self, json_obj, line_padding=""):
        # https://stackoverflow.com/questions/8988775/convert-json-to-xml-in-python
        result_list = list()

        json_obj_type = type(json_obj)

        if json_obj_type is list:
            for sub_elem in json_obj:
                result_list.append(self._json2xml(sub_elem, line_padding))
            return "\n".join(result_list)

        if json_obj_type is dict:
            for tag_name in json_obj:
                sub_obj = json_obj[tag_name]
                result_list.append("%s<%s>" % (line_padding, tag_name))
                result_list.append(self._json2xml(sub_obj, line_padding))
                result_list.append("%s</%s>" % (line_padding, tag_name))
            return "\n".join(result_list)

        return "%s%s" % (line_padding, json_obj)

    def invoke_tool(self, tool_name, return_results_only=False, **kwargs):
        try:
            cur_tool = self.tool_mapping.get(tool_name)
            # if 'error' in cur_tool.keys():
            #     return f'Error: {cur_tool["error"]}'
            if cur_tool is not None:
                t0 = time.time()
                if (
                    not hasattr(cur_tool, "requires_username")
                    or not cur_tool.requires_username
                ):
                    kwargs.pop("username", None)

                ans = cur_tool(**kwargs)
                self.invoke_log.append(
                    {
                        "tool_name": tool_name,
                        "execution_time": time.time() - t0,
                        "result_length": len(ans),
                    }
                )
                if return_results_only:
                    return ans
                else:
                    return self.call_return_string.replace(
                        "{{TOOLNAME}}", tool_name
                    ).replace("{{TOOLRESULTS}}", ans)
            else:
                return f"Tool {tool_name} not found. Please check tool name. Available tools: {self.tool_mapping.keys()}"
        except Exception as e:
            print(f"Tool execution failed: {str(e)}")
            return str(
                e
            )  # "Failed to invoke tool. Please try again, possibly in a different way."



# UNREVIEWED

import pytest
from unittest.mock import patch, mock_open
from gat_llm.tools.write_local_file import ToolWriteLocalFile


@pytest.fixture
def mock_os_path_isfile():
    with patch("os.path.isfile") as mock:
        yield mock


@pytest.fixture
def mock_os_makedirs():
    with patch("os.makedirs") as mock:
        yield mock


def test_unexpected_arg(unexpected_param_msg):
    twlf = ToolWriteLocalFile()
    ans = twlf("path/to/file.txt", "content", unexpected_argument=None)
    assert ans == f"{unexpected_param_msg}unexpected_argument"


def test_write_local_file_success(mock_os_path_isfile, mock_os_makedirs):
    mock_os_path_isfile.return_value = False
    twlf = ToolWriteLocalFile()
    with patch("builtins.open", mock_open()) as mock_file:
        result = twlf("path/to/file.txt", "Test content")
    assert "<outcome>" in result
    assert "File written successfully" in result
    assert "<path_to_file>path/to/file.txt</path_to_file>" in result
    assert "</outcome>" in result
    mock_file.assert_called_once_with("path/to/file.txt", "w", encoding="utf-8")
    mock_file().write.assert_called_once_with("Test content")


def test_write_local_file_already_exists(mock_os_path_isfile, mock_os_makedirs):
    mock_os_path_isfile.return_value = True
    twlf = ToolWriteLocalFile()
    result = twlf("path/to/existing_file.txt", "Test content")
    assert "<outcome>" in result
    assert "Error: File already exists" in result
    assert "This tool is not allowed to overwrite files" in result
    assert "</outcome>" in result


@pytest.mark.parametrize(
    "file_path,content",
    [
        ("media/new_file.txt", "Hello, World!"),
        ("path/to/document.md", "# Markdown Title\n\nContent here"),
        ("scripts/test.py", "print('Hello, Python!')"),
    ],
)
def test_write_local_file_various_types(
    file_path, content, mock_os_path_isfile, mock_os_makedirs
):
    mock_os_path_isfile.return_value = False
    twlf = ToolWriteLocalFile()
    with patch("builtins.open", mock_open()) as mock_file:
        result = twlf(file_path, content)
    assert "<outcome>" in result
    assert "File written successfully" in result
    assert f"<path_to_file>{file_path}</path_to_file>" in result
    assert "</outcome>" in result
    mock_file.assert_called_once_with(file_path, "w", encoding="utf-8")
    mock_file().write.assert_called_once_with(content)


# Add more tests for different scenarios and edge cases
